\ we use the garbage collector -- would this ever become a Scheme
\ without GC ;-)?
\ 64 1024 * 1024 * ( 64MB ) true false s" gc.fs" included drop 2drop

\ use this instead to disable GC (all memory stays allocated forever)
: alloc postpone allocate ; immediate

: scm-immediate? ( w -- f )
    ]] 1 literal and 0<> [[ ; immediate

\
\ inums: small integers
\

: scm-make-inum ( n -- w )
    2 lshift 3 or ; 

: scm-inum@ ( w -- n )
    2 rshift ;

: scm-inum? ( w -- f )
    3 and 3 = ; 



\
\ we can put objects that are very small into an immediate tagged with
\ a 4-bit typecode
\
: scm-tc4-instance? ( w -- f )
    ]] 3 and 3 = [[ ; immediate

: scm-tc4 ( w -- utc4 )
    ]] 3 rshift 15 and [[ ; immediate

: scm-tc4= ( w utc4 -- f )
    2 lshift swap scm-tc4 = ;

: scm-make-tc4-instance ( udata utc4 -- w )
    ]] 2 lshift 1 or swap
    6 lshift or [[ ; immediate

\
\ booleans
\
: scm-tc4-bool ( -- utc4 )
    ]] 0 [[ ; immediate

: scm-bool-f ( -- wbool )
    0 scm-tc4-bool scm-make-tc4-instance ;

: scm-bool-t ( -- wbool )
    -1 scm-tc4-bool scm-make-tc4-instance ;

: scm-bool@ ( wbool -- f )
    2 rshift 0<> ;

: scm-make-bool ( f -- wbool )
    if scm-bool-t
    else scm-bool-f
    endif ;

\
\ singular types
\
: scm-tc4-singular ( -- utc4 )
    ]] 1 [[ ; immediate

: scm-make-singular-instance ( utcs -- w )
    ]] scm-tc4-singular scm-make-tc4-instance [[ ; immediate

\ the magic unspecified object
: scm-unspecified ( -- w)
    ]] 0 scm-make-singular-instance [[ ; immediate

\ files
: scm-eof ( -- w )
    ]] 1 scm-make-singular-instance [[ ; immediate


\
\ generally, we use a separate forth cell for the type
\

: scm-tcw-alloc ( nchars ucells ucode -- a- )
    ]] swap 1+ cells alloc throw tuck ! [[ ; immediate

: scm-tcw-instance? ( w -- f )
    ]] 3 and 0= [[ ; immediate

: scm-:tcw-predicate ( tcw "name" -- )
    create , 
  does> ( w-cell -- )
    ( w-cell addr ) swap dup
    scm-tcw-instance? if @ swap @ =  else 2drop 0 endif ;



\
\ pairs
\

\ Note: We'd like to have cheap pairs (only two cells). Due to the GC,
\ we must not make the word that represents the pointer to the pair
\ point past the start of the pair memory, as normally done (by
\ setting one of the two least significant bits to 1 and masking these
\ two bits off when referencing the pair). Thus we can't have cheap
\ pairs unless we use another GC.

: scm-tcw-pair ( -- usize ucode )
    ]] 2 0 [[ ;  immediate

0 scm-:tcw-predicate scm-pair?

: scm-cons ( w-cdr w-car -- w )
    ]] scm-tcw-pair scm-tcw-alloc tuck cell+ ! tuck 2 cells + ! [[ ; immediate

: scm-pair@ ( w -- w-cdr w-car )
    ]] cell+ dup @ swap cell+ @ swap [[ ; immediate

: scm-car@ ( w -- w )
    ]] cell+ @ [[ ; immediate

: scm-cdr@ ( w -- w )
    ]] 2 cells + @ [[ ; immediate

: scm-car! ( w-car w -- )
    ]] cell+ ! [[ ; immediate
    
: scm-cdr! ( w-cdr w -- )
    ]] 2 cells + ! [[ ; immediate




\
\ procedures/lambdas
\
\ mem layout: utcw xt env
\
\ a procedure/lambda xt is supposed to have the following stack
\ effect: ( wargs wenv -- wresult )

: scm-tcw-procedure ( -- usize utcw )
    ]] 2 1 [[ ; immediate

1 scm-:tcw-predicate scm-procedure?

: scm-make-procedure ( wenv xt -- wproc )
    scm-tcw-procedure scm-tcw-alloc
    tuck cell+ !
    tuck 2 cells + ! ;

: scm-procedure@ ( wproc -- wenv xt )
    cell+ dup @ swap cell+ @ swap ;

: scm-procedure-apply ( wargs wproc -- wresult )
    scm-procedure@ >r swap r> execute ;


\
\ macros
\
\ mem layout: utcw xt env
\
\ a macro xt is supposed to have the following stack effect:
\ ( wargs wform -- wresult )
\
\ in fact, these are nearly the same as procedures, but are called on
\ parsed input, not evaluated arguments.

: scm-tcw-macro ( -- usize utcw )
    ]] 2 2 [[ ; immediate

2 scm-:tcw-predicate scm-macro?

: scm-make-macro ( wenv xt -- wproc )
    scm-tcw-macro scm-tcw-alloc
    tuck cell+ !
    tuck 2 cells + ! ;

: scm-macro@ ( wproc -- wenv xt )
    cell+ dup @ swap cell+ @ swap ;

: scm-macro-apply ( wargs wproc -- wresult )
    scm-macro@ nip execute ;

\
\ variables
\
\ mem layout: u-tc w-cell

: scm-tcw-variable ( usize u-tc )
    ]] 1 3 [[ ; immediate

3 scm-:tcw-predicate scm-variable?

: scm-make-variable ( w-init -- w )
    scm-tcw-variable scm-tcw-alloc tuck cell+ ! ;

: scm-variable@ ( w-var -- w )
    ]] cell+ @ [[ ; immediate

: scm-variable! ( w w-var -- )
    ]] cell+ ! [[ ; immediate

\
\ symbols
\
\ mem layout: u-tc u-nchars c-addr

: scm-tcw-symbol ( usize u-tc )
    ]] 2 4 [[ ; immediate

4 scm-:tcw-predicate scm-symbol?

\ table of interned symbols; xts push symbol cell
table constant scm-%sym-table

: scm-%make-symbol ( c-addr u-nchars -- w-sym )
    2dup 2>r scm-tcw-symbol scm-tcw-alloc tuck cell+ !
    tuck 2 cells + ! ( w-sym )
    2r> nextname scm-%sym-table >order definitions
    dup create ,
    previous definitions ;

: scm-%sym-lookup ( c-addr u-nchars -- { w-sym | w-bool-f } )
    scm-%sym-table search-wordlist 0= if
	scm-bool-f
    else
	execute @
    endif ;

: scm-make-symbol ( c-addr u-nchars -- w-sym )
    2dup scm-%sym-lookup dup scm-bool-f = if \ new symbol
	drop here over allot align swap 2dup 2>r move
	2r> scm-%make-symbol
    else
	nip nip
    endif ;


: scm-symbol@ ( w-sym -- c-addr u-nchars )
    cell+ dup @ swap cell+ @ swap ;

\
\
\ environments

\ mem layout: utcw wid wparent

\ implementation notes: we use chained tables to store the
\ bindings. The xt's in the table are supposed to push the address of
\ a w-var onto the stack.

: scm-tcw-environment ( -- usize utcw )
    ]] 2 5 [[ ; immediate

: scm-make-environment ( wparent -- w )
    scm-tcw-environment scm-tcw-alloc 
    dup cell+ table swap !
    tuck 2 cells + ! ;

: scm-environment@ ( w-env -- wparent wid )
    cell+ dup @ swap cell+ @ swap ;

: scm-%env-lookup ( w-env c-addr u-nchars -- w-parent 0 | xt +-1 )
    2>r scm-environment@ 2r> rot search-wordlist ;

: scm-%env-recurse? ( w-env c-addr u-nchars --
    { w-bool-f 2 | w-env c-addr u-nchars 0 } )
    rot dup scm-bool-f = if
	nip nip 2
    else
	-rot 0
    endif ;

\ look up recursivly the binding for a name. 
: scm-environment-lookup ( w-sym w-env -- w )
    swap scm-symbol@ 2dup 2>r
    0 begin
	drop 
	scm-%env-lookup
	dup 0= if drop ( w-parent )
	    2r@ scm-%env-recurse? 
	else
	    rot drop ( xt +-1 )
	endif
    dup 0<> until
    2rdrop
    case
	-1 of execute @ endof
    endcase ;

: scm-environment-define! ( w-var w-sym w-env -- )
    swap scm-symbol@ nextname ( w-var w-env )
    scm-environment@ nip >order definitions ( w-var )
    create ,
    previous definitions ;



\
\ compound data types
\

\
\ lists
\
: scm-eol ( -- w )
    ]] 1 scm-make-singular-instance [[ ; immediate

: scm-list-length ( wlist -- u )
    0 begin
	swap dup scm-eol <> while
	    scm-cdr@ swap 1+
    repeat drop ;

\ Note that the xt can refer to elements already on the stack
: scm-for-each ( wlist xt -- )
    swap begin
	dup scm-eol <> while
	    scm-pair@ ( xt cdr car ) swap >r swap dup >r execute r> r> 
    repeat 2drop ;

: scm.list ( wenv wargs -- wlist )
    nip ;

: scm-map ( wlist xt -- wlist )
    >r :noname r> ]] swap >r [[ compile, ( wtail wresult )
    ]] scm-eol swap scm-cons dup r> scm-cdr! ; [[
    scm-eol scm-bool-f scm-cons dup >r -rot scm-for-each drop r> scm-cdr@ ;

: scm.map ( wenv wargs -- wresult )
    nip \ environment not needed
    dup scm-list-length 2 <> s" map: invalid number of arguments" exception and throw
    scm-pair@ dup scm-procedure? 0=
    s" map: arg 1 not a procedure" exception and throw
    swap scm-car@ dup scm-pair? 0= s" map: arg 2 not a list" exception and throw
    ( wproc wlist ) swap
    >r :noname r> ]] literal swap scm-eol swap scm-cons
    swap scm-procedure-apply ; [[ scm-map ;

: scm.for-each ( wenv args -- wresult )
    \ not really efficient, but simple
    scm.map drop scm-unspecified ;

\
\ assertations
\
: scm-assert-inum ( w-inum -- w-inum )
    dup scm-inum? 0= s" inum expected" exception and throw ;


: scm-error ( c-addr nchars -- )
    exception throw ;

\ display
defer scm-display

:noname ( obj -- ) { obj }
    obj scm-symbol? if
	obj scm-symbol@ type exit
    endif
    obj scm-inum? if
	obj scm-inum@ . exit
    endif
    obj scm-pair? if
	s" (" type obj ['] scm-display scm-for-each s" )" type exit
    endif
    obj scm-unspecified = if
	s" #<unspecified>" type exit
    endif
    obj scm-eof = if
	s" #<eof>" type exit
    endif
    obj scm-eol = if
	s" ()" type exit
    endif
    s" unknown object passed to display" scm-error ;
is scm-display

: scm.display ( wenv wargs -- wresult )
    nip
    dup scm-list-length 1 <> s" display: wrong number of args" exception and throw
    scm-car@ scm-display scm-unspecified ;
    
: scm.newline ( wenv wargs -- )
    nip
    dup scm-list-length 0 <> s" newline: wrong number of args" exception and throw
    cr ;




\ arch-tag: e5d760d9-2359-4b31-af6b-c76acfaa1333
\ (do not remove this comment)
